"""Project build script to convert this project's AsciiDoc README file to Markdown format."""

import inspect
import os
import shutil
import subprocess
import sys
import warnings
from collections.abc import Collection, Iterable
from pathlib import Path
from typing import TYPE_CHECKING

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

from hatchling.metadata.plugin.interface import MetadataHookInterface

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from typing import Final, LiteralString

__all__: "Sequence[str]" = ("PydowndocCustomReadmeMetadataHook",)


class PydowndocCustomReadmeMetadataHook(MetadataHookInterface):
    """Hatchling metadata hook to convert this project's AsciiDoc file to Markdown format."""

    @classmethod
    def _get_readme_path(cls, config: "Mapping[str, object]", root: "Path") -> "Path":
        readme_path: str = ""

        config_name: LiteralString
        for config_name in ("readme-path", "path"):
            try:
                raw_readme_path: object = config[config_name]
            except KeyError:
                continue

            if not isinstance(raw_readme_path, str):
                INVALID_PATH_TYPE_MESSAGE: str = (
                    f"{cls.PLUGIN_NAME}.{config_name} must be a string."
                )
                raise TypeError(INVALID_PATH_TYPE_MESSAGE)

            readme_path = raw_readme_path
            break

        return root / (readme_path if readme_path else "README.adoc")

    @classmethod
    def _is_project_misconfigured(cls, metadata: "Mapping[str, object]") -> bool:
        if "readme" in metadata:
            return True

        dynamic: "object | Collection[object]" = metadata.get("dynamic", [])
        if not isinstance(dynamic, Collection):
            INVALID_DYNAMIC_TYPE_MESSAGE: Final[str] = (
                "'dynamic' field within `[project]` must be an array."
            )
            raise TypeError(INVALID_DYNAMIC_TYPE_MESSAGE)

        return "readme" not in dynamic

    @classmethod
    def _perform_conversion(cls, readme_path: "Path") -> str:
        downdoc_executable: "str | None" = shutil.which("downdoc")
        if downdoc_executable is None:
            DOWNDOC_NOT_INSTALLED_MESSAGE: Final[str] = (
                "The downdoc executable could not be found."
            )
            raise OSError(DOWNDOC_NOT_INSTALLED_MESSAGE)

        return subprocess.run(
            (downdoc_executable, "--output", "-", "--", str(readme_path)),
            capture_output=True,
            text=True,
            check=True,
        ).stdout

    @override
    def update(self, metadata: dict[str, object]) -> None:
        if (
            ("update", __file__)
            in ((frame.function, frame.filename) for frame in inspect.stack()[1:])
        ):  # SOURCE: https://github.com/flying-sheep/hatch-docstring-description/blob/2dfbfba2c48e112825fdd0cb7c37035d5598224c/src/hatch_docstring_description/read_description.py#L21
            return

        if self._is_project_misconfigured(metadata):
            MISSING_DYNAMIC_MESSAGE: Final[str] = (
                "You must add 'readme' to your `dynamic` fields and not to `[project]`."
            )
            raise TypeError(MISSING_DYNAMIC_MESSAGE)

        readme_path: Path = self._get_readme_path(self.config, Path(self.root))

        if not readme_path.is_file():
            raise FileNotFoundError(str(readme_path))

        try:
            metadata["readme"] = {
                "content-type": "text/markdown",
                "text": self._perform_conversion(readme_path),
            }
        except OSError as e:
            if (
                "/renovate/" not in os.environ["PWD"]
                and os.getenv("SKIP_MISSING_DOWNDOC", "False") != "True"
            ):
                raise e from e

            warnings.warn(
                (
                    f"{e} "
                    "This package will be built without any README content, "
                    "it MUST NOT BE UPLOADED to any package distribution platform "
                    "(E.g. PyPI)."
                ),
                stacklevel=1,
            )
            metadata["readme"] = {
                "content-type": "text/plain",
                "text": (
                    "Missing README content. "
                    "DO NOT UPLOAD this package to any distribution platform (E.g. PyPI).\n\n"
                    "If you are seeing this message on a package distribution platform, "
                    "please contact the project's maintainer."
                ),
            }

        if isinstance(metadata["dynamic"], Iterable):
            metadata["dynamic"] = [value for value in metadata["dynamic"] if value != "readme"]



"""Python wrapper for converting/reducing AsciiDoc files back to Markdown."""

from typing import TYPE_CHECKING, overload

from ._utils import OUTPUT_CONVERSION_TO_STRING, ConversionError
from .conversion_backends import DowndocMarkdownConversionBackend

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from pathlib import Path

    from ._utils import ConversionOutputDestinationFlag
    from .conversion_backends import BaseConversionBackend

__all__: "Sequence[str]" = (
    "OUTPUT_CONVERSION_TO_STRING",
    "ConversionError",
    "convert_file",
    "convert_string",
    "get_version",
)


def get_version(
    backend: "type[BaseConversionBackend]" = DowndocMarkdownConversionBackend,
) -> str:
    """
    Retrieve the current version of the given conversion backend.

    Arguments:
        backend: The conversion backend to get the version of, defaults to `DOWNDOC_MD`.

    Returns:
        Conversion backend's version text output.

    Raises:
        subprocess.CalledProcessError: If calling the conversion backend subprocess exited
            with a non-zero exit code.
    """
    return backend.get_version()


@overload
def convert_file(
    file_path: "Path",
    *,
    output_location: "ConversionOutputDestinationFlag",
    backend: "type[BaseConversionBackend]" = ...,
    attributes: "Mapping[str, str] | None" = ...,
    postpublish: bool = ...,
    prepublish: bool = ...,
) -> str: ...


@overload
def convert_file(
    file_path: "Path",
    *,
    attributes: "Mapping[str, str] | None" = ...,
    output_location: "Path | None" = ...,
    backend: "type[BaseConversionBackend]" = ...,
    postpublish: bool = ...,
    prepublish: bool = ...,
) -> None: ...


def convert_file(
    file_path: "Path",
    *,
    attributes: "Mapping[str, str] | None" = None,
    output_location: "Path | ConversionOutputDestinationFlag | None" = None,
    backend: "type[BaseConversionBackend]" = DowndocMarkdownConversionBackend,
    postpublish: bool = False,
    prepublish: bool = False,
) -> "str | None":
    """
    Execute the downdoc converter upon the given input file path.

    Arguments:
        file_path: The location of the file to convert from AsciiDoc to Markdown.
        attributes: AsciiDoc attributes to be set while rendering AsciiDoc files.
        output_location: The location to save the converted Markdown output,
            or `OUTPUT_CONVERSION_TO_STRING` to return as a string.
            By default (or when `None`), the output file will use the same name
            as the input file, with the extension changed to `.md`.
        backend: The conversion backend to use, defaults to `downdoc-md`.
        postpublish: Whether to run the postpublish lifecycle routine (restore the input file).
        prepublish: Whether to run the prepublish lifecycle routine
            (convert and hide the input file).

    Returns:
        `None`, or the converted Markdown output
        when `output_location` is `OUTPUT_CONVERSION_TO_STRING`.

    Raises:
        ConversionError: When calling the downdoc subprocess exited with an error.
    """
    return backend.convert_file(
        file_path=file_path,
        attributes=attributes,
        output_location=output_location,
        postpublish=postpublish,
        prepublish=prepublish,
    )


def convert_string(
    asciidoc_content: str,
    *,
    attributes: "Mapping[str, str] | None" = None,
    backend: "type[BaseConversionBackend]" = DowndocMarkdownConversionBackend,
) -> str:
    """
    Execute the downdoc converter upon the given AsciiDoc content string.

    Arguments:
        asciidoc_content: The string AsciiDoc content to convert.
        attributes: AsciiDoc attributes to be set while rendering AsciiDoc files.
        backend: The conversion backend to use, defaults to `downdoc-md`.

    Returns:
        The converted Markdown output.

    Raises:
        ConversionError: When calling the downdoc subprocess exited with an error.
    """
    return backend.convert_string(asciidoc_content=asciidoc_content, attributes=attributes)



"""
Hatch build package hooks entry-points.

Convert your AsciiDoc README to the Markdown format, supported by PyPA's README specification:
https://packaging.python.org/en/latest/specifications/pyproject-toml/#readme
"""

import inspect
import sys
from collections.abc import Collection, Iterable
from pathlib import Path
from typing import TYPE_CHECKING

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

from hatchling.metadata.plugin.interface import MetadataHookInterface
from hatchling.plugin import hookimpl
from typed_classproperties import classproperty

import pydowndoc
from pydowndoc.conversion_backends import (
    DowndocMarkdownConversionBackend,
    PandocMarkdownConversionBackend,
    PandocMultiMarkdownConversionBackend,
    PandocPHPMarkdownExtraConversionBackend,
    PandocRSTConversionBackend,
    PandocTXTConversionBackend,
)

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from typing import Final

    from pydowndoc.conversion_backends import BaseConversionBackend

__all__: "Sequence[str]" = ("DowndocReadmeMetadataHook", "hatch_register_metadata_hook")


class DowndocReadmeMetadataHook(MetadataHookInterface):
    """Hatchling metadata hook for converting AsciiDoc README files to Markdown format."""

    @classproperty
    @override
    def PLUGIN_NAME(cls) -> str:
        return "downdoc-readme"

    @classmethod
    def _get_readme_path(cls, config: "Mapping[str, object]", root: Path) -> Path:
        raw_readme_path: "object | str" = config.get("path", "")

        if not isinstance(raw_readme_path, str):
            INVALID_PATH_TYPE_MESSAGE: Final[str] = f"{cls.PLUGIN_NAME}.path must be a string."
            raise TypeError(INVALID_PATH_TYPE_MESSAGE)

        return root / (raw_readme_path if raw_readme_path else "README.adoc")

    @classmethod
    def _get_conversion_backend(
        cls, config: "Mapping[str, object]"
    ) -> "type[BaseConversionBackend] | None":
        raw_conversion_backend: "object | str" = config.get("backend", "")

        if not isinstance(raw_conversion_backend, str):
            INVALID_CONVERSION_BACKEND_TYPE_MESSAGE: Final[str] = (
                f"{cls.PLUGIN_NAME}.backend must be a string."
            )
            raise TypeError(INVALID_CONVERSION_BACKEND_TYPE_MESSAGE)

        raw_conversion_backend = raw_conversion_backend.strip()

        if not raw_conversion_backend:
            return None

        try:
            return {
                conversion_backend.ID: conversion_backend
                for conversion_backend in (
                    DowndocMarkdownConversionBackend,
                    PandocMarkdownConversionBackend,
                    PandocMultiMarkdownConversionBackend,
                    PandocPHPMarkdownExtraConversionBackend,
                    PandocTXTConversionBackend,
                    PandocRSTConversionBackend,
                )
            }[raw_conversion_backend]
        except KeyError as e:
            INVALID_CONVERSION_BACKEND_MESSAGE: Final[str] = (
                f"Invalid conversion_backend: {raw_conversion_backend!r}."
            )
            raise ValueError(INVALID_CONVERSION_BACKEND_MESSAGE) from e

    @classmethod
    def _is_project_misconfigured(cls, metadata: "Mapping[str, object]") -> bool:
        if "readme" in metadata:
            return True

        dynamic: "object | Collection[object]" = metadata.get("dynamic", [])
        if not isinstance(dynamic, Collection):
            INVALID_DYNAMIC_TYPE_MESSAGE: Final[str] = (
                "'dynamic' field within `[project]` must be an array."
            )
            raise TypeError(INVALID_DYNAMIC_TYPE_MESSAGE)

        return "readme" not in dynamic

    @override
    def update(self, metadata: dict[str, object]) -> None:
        if (
            ("update", __file__)
            in ((frame.function, frame.filename) for frame in inspect.stack()[1:])
        ):  # SOURCE: https://github.com/flying-sheep/hatch-docstring-description/blob/2dfbfba2c48e112825fdd0cb7c37035d5598224c/src/hatch_docstring_description/read_description.py#L21
            return

        if self._is_project_misconfigured(metadata):
            MISSING_DYNAMIC_MESSAGE: Final[str] = (
                "You must add 'readme' to your `dynamic` fields and not to `[project]`."
            )
            raise TypeError(MISSING_DYNAMIC_MESSAGE)

        readme_path: Path = self._get_readme_path(self.config, Path(self.root))

        if not readme_path.is_file():
            raise FileNotFoundError(str(readme_path))

        conversion_backend: "type[BaseConversionBackend] | None" = (
            self._get_conversion_backend(self.config)
        )

        metadata["readme"] = {
            "content-type": "text/markdown",
            "text": (
                pydowndoc.convert_file(
                    readme_path,
                    output_location=pydowndoc.OUTPUT_CONVERSION_TO_STRING,
                    backend=conversion_backend,
                )
                if conversion_backend is not None
                else pydowndoc.convert_file(
                    readme_path,
                    output_location=pydowndoc.OUTPUT_CONVERSION_TO_STRING,
                )
            ),
        }

        if isinstance(metadata["dynamic"], Iterable):
            metadata["dynamic"] = [value for value in metadata["dynamic"] if value != "readme"]


@hookimpl
def hatch_register_metadata_hook() -> (
    "type[MetadataHookInterface] | list[type[MetadataHookInterface]]"
):
    """Export the correct metadata hook class for hatch projects."""
    return DowndocReadmeMetadataHook



"""Common utility classes internal to this project."""

import sys
from typing import TYPE_CHECKING

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Final

__all__: "Sequence[str]" = (
    "OUTPUT_CONVERSION_TO_STRING",
    "ConversionError",
    "ConversionOutputDestinationFlag",
)


class ConversionOutputDestinationFlag:
    pass


OUTPUT_CONVERSION_TO_STRING: "Final[ConversionOutputDestinationFlag]" = (
    ConversionOutputDestinationFlag()
)


class ConversionError(RuntimeError):
    """Raised when an error occurs while using the downdoc binary in a subprocess."""

    @override
    def __init__(
        self,
        message: "str | None" = None,
        *,
        subprocess_return_code: "int | None" = None,
        subprocess_stderr: "str | None" = None,
    ) -> None:
        self.message: "str | None" = message
        self.subprocess_return_code: "int | None" = subprocess_return_code
        self.subprocess_stderr: "str | None" = subprocess_stderr



"""Conversion backend classes relating to alternative executable conversion programs."""

import abc
import itertools
import shlex
import shutil
import subprocess
import sys
from pathlib import Path
from typing import TYPE_CHECKING, final, overload

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

from typed_classproperties import classproperty

from ._utils import OUTPUT_CONVERSION_TO_STRING

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence
    from typing import Final, NoReturn

    if sys.version_info >= (3, 11):
        from typing import LiteralString
    else:
        from typing_extensions import LiteralString

    from ._utils import ConversionOutputDestinationFlag

__all__: "Sequence[str]" = (
    "BaseConversionBackend",
    "DowndocMarkdownConversionBackend",
    "PandocMarkdownConversionBackend",
    "PandocMultiMarkdownConversionBackend",
    "PandocPHPMarkdownExtraConversionBackend",
    "PandocRSTConversionBackend",
    "PandocTXTConversionBackend",
)


class BaseConversionBackend(abc.ABC):
    """Abstract base class defining the structural API of a conversion backend."""

    @classproperty
    @abc.abstractmethod
    def ID(cls) -> "LiteralString":  # noqa: D102, N802
        pass

    @classmethod
    @abc.abstractmethod
    def get_version(cls) -> str:
        """Retrieve the version of the executable used by this conversion backend."""

    @classmethod
    @abc.abstractmethod
    def _convert_string(
        cls, asciidoc_content: str, *, attributes: "Mapping[str, str] | None" = None
    ) -> str:
        pass

    @final
    @classmethod
    def convert_string(
        cls, asciidoc_content: str, *, attributes: "Mapping[str, str] | None" = None
    ) -> str:
        """Convert AsciiDoc string content to an interpretable documentation format."""
        if not asciidoc_content.strip():
            INVALID_ASCIIDOC_CONTENT_MESSAGE: Final[str] = (
                "Cannot convert empty string content."
            )
            raise ValueError(INVALID_ASCIIDOC_CONTENT_MESSAGE)

        return cls._convert_string(asciidoc_content=asciidoc_content, attributes=attributes)

    @overload
    @classmethod
    @abc.abstractmethod
    def _convert_file(
        cls,
        file_path: "Path",
        *,
        output_location: "ConversionOutputDestinationFlag",
        attributes: "Mapping[str, str] | None" = ...,
        postpublish: bool = ...,
        prepublish: bool = ...,
    ) -> str: ...

    @overload
    @classmethod
    @abc.abstractmethod
    def _convert_file(
        cls,
        file_path: "Path",
        *,
        attributes: "Mapping[str, str] | None" = ...,
        output_location: "Path | None" = ...,
        postpublish: bool = ...,
        prepublish: bool = ...,
    ) -> None: ...

    @classmethod
    @abc.abstractmethod
    def _convert_file(
        cls,
        file_path: "Path",
        *,
        attributes: "Mapping[str, str] | None" = None,
        output_location: "Path | ConversionOutputDestinationFlag | None" = None,
        postpublish: bool = False,
        prepublish: bool = False,
    ) -> "str | None":
        pass

    @overload
    @classmethod
    def convert_file(
        cls,
        file_path: "Path",
        *,
        output_location: "ConversionOutputDestinationFlag",
        attributes: "Mapping[str, str] | None" = ...,
        postpublish: bool = ...,
        prepublish: bool = ...,
    ) -> str: ...

    @overload
    @classmethod
    def convert_file(
        cls,
        file_path: "Path",
        *,
        attributes: "Mapping[str, str] | None" = ...,
        output_location: "Path | None" = ...,
        postpublish: bool = ...,
        prepublish: bool = ...,
    ) -> None: ...

    @final
    @classmethod
    def convert_file(
        cls,
        file_path: "Path",
        *,
        attributes: "Mapping[str, str] | None" = None,
        output_location: "Path | ConversionOutputDestinationFlag | None" = None,
        postpublish: bool = False,
        prepublish: bool = False,
    ) -> "str | None":
        """Convert an AsciiDoc file to an interpretable documentation format."""
        if not file_path.is_file():
            raise FileNotFoundError(file_path)

        return cls._convert_file(
            file_path=file_path,
            attributes=attributes,
            output_location=output_location,
            postpublish=postpublish,
            prepublish=prepublish,
        )

    @classmethod
    def _attributes_to_arguments(
        cls, attributes: "Mapping[str, str] | None"
    ) -> "Iterable[str]":
        if attributes is None:
            attributes = {}

        return itertools.chain.from_iterable(
            ("--attribute", f"{shlex.quote(name)}={shlex.quote(val)}")
            for name, val in attributes.items()
        )

    if not TYPE_CHECKING:

        @override
        def __new__(cls) -> "NoReturn":
            CANNOT_INSTANTIATE_OBJECTS_MESSAGE: Final[str] = (
                "Cannot instantiate objects of this type."
            )
            raise RuntimeError(CANNOT_INSTANTIATE_OBJECTS_MESSAGE)


class DowndocMarkdownConversionBackend(BaseConversionBackend):
    """Backend to convert AsciiDoc content to Markdown using downdoc."""

    @classproperty
    @override
    def ID(cls) -> "LiteralString":
        return "downdoc-md"

    @classmethod
    def _get_downdoc_executable_path(cls) -> str:
        downdoc_executable: "str | None" = shutil.which("downdoc")

        if downdoc_executable is None:
            DOWNDOC_NOT_INSTALLED_MESSAGE: Final[str] = (
                "The downdoc executable could not be found. "
                "Ensure it is installed (E.g `uv add Pydowndoc[bin]`)."
            )
            raise OSError(DOWNDOC_NOT_INSTALLED_MESSAGE)

        return downdoc_executable

    @classmethod
    @override
    def get_version(cls) -> str:
        return subprocess.run(
            (cls._get_downdoc_executable_path(), "--version"),
            check=True,
            text=True,
            capture_output=True,
        ).stdout.strip()

    @classmethod
    @override
    def _convert_string(
        cls, asciidoc_content: str, *, attributes: "Mapping[str, str] | None" = None
    ) -> str:
        ends_with_newline: bool = asciidoc_content.endswith("\n")

        converted_string: str = subprocess.run(
            (
                cls._get_downdoc_executable_path(),
                *cls._attributes_to_arguments(attributes),
                "--output",
                "-",
                "--",
                "-",
            ),
            check=True,
            input=asciidoc_content,
            text=True,
            capture_output=True,
        ).stdout

        return converted_string if ends_with_newline else converted_string.removesuffix("\n")

    @overload
    @classmethod
    @override
    def _convert_file(
        cls,
        file_path: "Path",
        *,
        output_location: "ConversionOutputDestinationFlag",
        attributes: "Mapping[str, str] | None" = ...,
        postpublish: bool = ...,
        prepublish: bool = ...,
    ) -> str: ...

    @overload
    @classmethod
    @override
    def _convert_file(
        cls,
        file_path: "Path",
        *,
        attributes: "Mapping[str, str] | None" = ...,
        output_location: "Path | None" = ...,
        postpublish: bool = ...,
        prepublish: bool = ...,
    ) -> None: ...

    @classmethod
    @override
    def _convert_file(
        cls,
        file_path: "Path",
        *,
        attributes: "Mapping[str, str] | None" = None,
        output_location: "Path | ConversionOutputDestinationFlag | None" = None,
        postpublish: bool = False,
        prepublish: bool = False,
    ) -> "str | None":
        optional_arguments: list[str] = []

        if output_location == OUTPUT_CONVERSION_TO_STRING:
            optional_arguments.extend(("--output", "-"))
        elif isinstance(output_location, Path):
            optional_arguments.extend(("--output", str(output_location)))

        if postpublish:
            optional_arguments.extend("--postpublish")
        if prepublish:
            optional_arguments.extend("--prepublish")

        subprocess_stdout: str = subprocess.run(
            (
                cls._get_downdoc_executable_path(),
                *cls._attributes_to_arguments(attributes),
                *optional_arguments,
                "--",
                str(file_path),
            ),
            check=True,
            text=True,
            capture_output=True,
        ).stdout

        return subprocess_stdout if output_location == OUTPUT_CONVERSION_TO_STRING else None


class _BasePandocConversionBackend(BaseConversionBackend, abc.ABC):
    @classproperty
    @abc.abstractmethod
    def PANDOC_ID(cls) -> "LiteralString":  # noqa: N802
        pass

    @classproperty
    @abc.abstractmethod
    def FILE_SUFFIX(cls) -> "LiteralString":  # noqa: N802
        pass

    @classmethod
    def _get_asciidoctor_executable_path(cls) -> str:
        asciidoctor_executable: "str | None" = shutil.which("asciidoctor")

        if asciidoctor_executable is None:
            ASCIIDOCTOR_NOT_INSTALLED_MESSAGE: Final[str] = (
                "The asciidoctor executable could not be found. "
                "Ensure it is installed (https://docs.asciidoctor.org/asciidoctor/latest/install)."
            )
            raise OSError(ASCIIDOCTOR_NOT_INSTALLED_MESSAGE)

        return asciidoctor_executable

    @classmethod
    def _get_pandoc_executable_path(cls) -> str:
        pandoc_executable: "str | None" = shutil.which("pandoc")

        if pandoc_executable is None:
            PANDOC_NOT_INSTALLED_MESSAGE: Final[str] = (
                "The pandoc executable could not be found. "
                "Ensure it is installed (https://pandoc.org/installing.html)."
            )
            raise OSError(PANDOC_NOT_INSTALLED_MESSAGE)

        return pandoc_executable

    @classmethod
    @override
    def get_version(cls) -> str:
        # NOTE: Intermediary variables are used to prevent 'Unterminated f-string literal' errors in older Python versions
        asciidoctor_version: str = subprocess.run(
            (cls._get_asciidoctor_executable_path(), "--version"),
            check=True,
            text=True,
            capture_output=True,
        ).stdout.strip()

        pandoc_version: str = subprocess.run(
            (cls._get_pandoc_executable_path(), "--version"),
            check=True,
            text=True,
            capture_output=True,
        ).stdout.strip()

        return f"{asciidoctor_version}\n{pandoc_version}"

    @classmethod
    @override
    def _convert_string(
        cls, asciidoc_content: str, *, attributes: "Mapping[str, str] | None" = None
    ) -> str:
        return subprocess.run(
            (
                cls._get_pandoc_executable_path(),
                "--from",
                "docbook",
                "--to",
                cls.PANDOC_ID,
                "--output",
                "-",
                "--fail-if-warnings",
            ),
            check=True,
            input=(
                subprocess.run(
                    (
                        cls._get_asciidoctor_executable_path(),
                        *cls._attributes_to_arguments(attributes),
                        "--out-file",
                        "-",
                        "--backend",
                        "docbook5",
                        "--warnings",
                        "--failure-level",
                        "WARNING",
                        "--",
                        "-",
                    ),
                    check=True,
                    input=asciidoc_content,
                    text=True,
                    capture_output=True,
                ).stdout
            ),
            text=True,
            capture_output=True,
        ).stdout

    @overload
    @classmethod
    @override
    def _convert_file(
        cls,
        file_path: "Path",
        *,
        output_location: "ConversionOutputDestinationFlag",
        attributes: "Mapping[str, str] | None" = ...,
        postpublish: bool = ...,
        prepublish: bool = ...,
    ) -> str: ...

    @overload
    @classmethod
    @override
    def _convert_file(
        cls,
        file_path: "Path",
        *,
        attributes: "Mapping[str, str] | None" = ...,
        output_location: "Path | None" = ...,
        postpublish: bool = ...,
        prepublish: bool = ...,
    ) -> None: ...

    @classmethod
    @override
    def _convert_file(
        cls,
        file_path: "Path",
        *,
        attributes: "Mapping[str, str] | None" = None,
        output_location: "Path | ConversionOutputDestinationFlag | None" = None,
        postpublish: bool = False,
        prepublish: bool = False,
    ) -> "str | None":
        if postpublish or prepublish:
            INVALID_PREPUBLISH_OR_POSTPUBLISH_MESSAGE: Final[str] = (
                "Neither 'postpublish' nor 'prepublish' can be used "
                "for this conversion backend."
            )
            raise ValueError(INVALID_PREPUBLISH_OR_POSTPUBLISH_MESSAGE)

        optional_arguments: list[str] = []

        if output_location is None:
            optional_arguments.extend(
                ("--output", str(file_path.with_suffix(cls.FILE_SUFFIX)))
            )
        elif output_location == OUTPUT_CONVERSION_TO_STRING:
            optional_arguments.extend(("--output", "-"))
        elif isinstance(output_location, Path):
            optional_arguments.extend(("--output", str(output_location)))

        subprocess_stdout: str = subprocess.run(
            (
                cls._get_pandoc_executable_path(),
                "--from",
                "docbook",
                "--to",
                cls.PANDOC_ID,
                *optional_arguments,
                "--fail-if-warnings",
            ),
            check=True,
            input=(
                subprocess.run(
                    (
                        cls._get_asciidoctor_executable_path(),
                        *cls._attributes_to_arguments(attributes),
                        "--out-file",
                        "-",
                        "--backend",
                        "docbook5",
                        "--warnings",
                        "--failure-level",
                        "WARNING",
                        "--",
                        str(file_path),
                    ),
                    check=True,
                    text=True,
                    capture_output=True,
                ).stdout
            ),
            text=True,
            capture_output=True,
        ).stdout

        return subprocess_stdout if output_location == OUTPUT_CONVERSION_TO_STRING else None


class PandocMarkdownConversionBackend(_BasePandocConversionBackend):
    """Backend to convert AsciiDoc content to Pandoc's Markdown using pandoc & Asciidoctor."""

    @classproperty
    @override
    def ID(cls) -> "LiteralString":
        return "pandoc-md"

    @classproperty
    @override
    def PANDOC_ID(cls) -> "LiteralString":
        return "markdown"

    @classproperty
    @override
    def FILE_SUFFIX(cls) -> "LiteralString":
        return ".md"


class PandocMultiMarkdownConversionBackend(_BasePandocConversionBackend):
    """Backend to convert AsciiDoc content to MultiMarkdown using pandoc & Asciidoctor."""

    @classproperty
    @override
    def ID(cls) -> "LiteralString":
        return "pandoc-multi-md"

    @classproperty
    @override
    def PANDOC_ID(cls) -> "LiteralString":
        return "markdown_mmd"

    @classproperty
    @override
    def FILE_SUFFIX(cls) -> "LiteralString":
        return ".md"


class PandocPHPMarkdownExtraConversionBackend(_BasePandocConversionBackend):
    """Backend to convert AsciiDoc content to PHP Markdown Extra using pandoc & Asciidoctor."""

    @classproperty
    @override
    def ID(cls) -> "LiteralString":
        return "pandoc-php-md-extra"

    @classproperty
    @override
    def PANDOC_ID(cls) -> "LiteralString":
        return "markdown_phpextra"

    @classproperty
    @override
    def FILE_SUFFIX(cls) -> "LiteralString":
        return ".md"


class PandocTXTConversionBackend(_BasePandocConversionBackend):
    """Backend to convert AsciiDoc content to plaintext using pandoc & Asciidoctor."""

    @classproperty
    @override
    def ID(cls) -> "LiteralString":
        return "pandoc-txt"

    @classproperty
    @override
    def PANDOC_ID(cls) -> "LiteralString":
        return "plain"

    @classproperty
    @override
    def FILE_SUFFIX(cls) -> "LiteralString":
        return ".txt"


class PandocRSTConversionBackend(_BasePandocConversionBackend):
    """Backend to convert AsciiDoc content to reStructuredText using pandoc & Asciidoctor."""

    @classproperty
    @override
    def ID(cls) -> "LiteralString":
        return "pandoc-rst"

    @classproperty
    @override
    def PANDOC_ID(cls) -> "LiteralString":
        return "rst"

    @classproperty
    @override
    def FILE_SUFFIX(cls) -> "LiteralString":
        return ".rst"
